package at.ac.uibk.gitsearch.service.vocabulary;

import at.ac.uibk.gitsearch.domain.vocabulary.VocabularyItem;
import at.ac.uibk.gitsearch.domain.vocabulary.VocabularyItem.LanguageItem;
import at.ac.uibk.gitsearch.repository.vocabulary.VocabularyRepository;
import at.ac.uibk.gitsearch.service.vocabulary.VocabularyService.VocabularyServiceConfig.RequiredEnum;
import at.ac.uibk.gitsearch.service.vocabulary.VocabularyService.VocabularyServiceConfig.VocabularySetting;
import com.fasterxml.jackson.annotation.JsonGetter;
import com.fasterxml.jackson.annotation.JsonSetter;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.MapperFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.json.JsonMapper;
import java.beans.IntrospectionException;
import java.beans.PropertyDescriptor;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import javax.annotation.PostConstruct;
import org.apache.lucene.search.spell.LevenshteinDistance;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO.InteractivityTypeXXX;
import org.codeability.sharing.plugins.api.search.UserProvidedMetadataDTO.Person;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

@Service
@Transactional
public class VocabularyService {

    private static final String PERSON_ARRAY_CLASSNAME = Person.class.getCanonicalName() + "[]";
    private final Logger log = LoggerFactory.getLogger(VocabularyService.class);
    private static final float MAXIMAL_DISTANCE = 0.70f;

    @Autowired
    private VocabularyRepository vocabularyRepository;

    private VocabularyServiceConfig config;

    @PostConstruct
    void loadConfig() throws IOException {
        config = readVocabularyConfig();
    }

    public Map<String, URI> getVocabularyURLMap() {
        return vocabularyRepository.getVocabularyURLMap();
    }

    /**
     * returns the vocabulary items for the field.
     *
     * @param field
     * @return
     */

    public VocabularyItem[] getVocabularyItemsFor(String field) {
        return vocabularyRepository.getVocabularyItemsFor(field);
    }

    public VocabularyItem[] getKeywordItems() {
        return vocabularyRepository.getVocabularyItemsFor("Keyword", VocabularyItem.class);
    }

    public VocabularyItem[] getProgrammingLanguageItems() {
        return vocabularyRepository.getVocabularyItemsFor("Programming Language", VocabularyItem.class);
    }

    public VocabularyItem[] getDifficultyItems() {
        return vocabularyRepository.getVocabularyItemsFor("Difficulty", VocabularyItem.class);
    }

    /**
     * validates the metadata
     *
     * @param metadata   the metadata to check
     * @param isTopLevel is metaData on topLevel
     * @return
     */
    public ValidationResultDTO validateMetaData(UserProvidedMetadataDTO metadata, boolean isTopLevel) {
        ValidationResultDTO result = new ValidationResultDTO();

        for (VocabularySetting setting : config.config) {
            try {
                checkField(
                    setting.getGetter().invoke(metadata),
                    setting.getProperty(),
                    setting.getOEResourceType(),
                    isTopLevel
                        ? (setting.getRequired() == RequiredEnum.REQUIRED_ON_TOP_LEVEL || setting.getRequired() == RequiredEnum.REQUIRED)
                        : setting.getRequired() == RequiredEnum.REQUIRED,
                    setting.isExactMatch(),
                    setting.getExtraEntries(),
                    result
                );
            } catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
                log.error("Cannot check field {}", setting.getProperty(), e);
            }
        }

        return result;
    }

    protected void checkField(
        Object fieldValue,
        String fieldName,
        String apiFieldName,
        boolean required,
        boolean exactMatch,
        String[] extraEntries,
        ValidationResultDTO result
    ) {
        if (fieldValue == null) {
            if (required) {
                result.errors.add(String.format("The field %s is mandatory", fieldName));
            }
            return;
        }

        final String[] values;
        if (fieldValue instanceof String) {
            values = new String[] { (String) fieldValue };
        } else if (fieldValue instanceof String[]) {
            values = (String[]) fieldValue;
        } else if (fieldValue instanceof InteractivityTypeXXX) {
            values = new String[] { ((InteractivityTypeXXX) fieldValue).getExternalName() };
        } else {
            if (PERSON_ARRAY_CLASSNAME.equals(fieldValue.getClass().getCanonicalName())) {
                log.debug("We cannot check field {} of type {}", apiFieldName, fieldValue.getClass().getCanonicalName());
            } else {
                log.error("Cannot check field {} of type {}", apiFieldName, fieldValue.getClass().getCanonicalName());
            }
            return;
        }
        checkField(values, fieldName, apiFieldName, required, exactMatch, extraEntries, result);
    }

    @SuppressWarnings({ "PMD.CyclomaticComplexity", "PMD.CognitiveComplexity" })
    protected void checkField(
        String[] values,
        String fieldName,
        String apiFieldName,
        boolean required,
        boolean exactMatch,
        String[] extraEntries,
        ValidationResultDTO result
    ) {
        if (required && (values.length == 0 || (values.length == 1 && values[0] == null))) {
            result.errors.add(String.format("The field %s is mandatory", fieldName));
            return;
        }
        specialFieldHandling(values, fieldName, result);
        if (apiFieldName != null) {
            final VocabularyItem[] fieldValues = getVocabularyItemsFor(apiFieldName);
            final List<String> notFoundValues = new ArrayList<>();
            for (String value : values) {
                boolean matchFound = Arrays
                    .stream(fieldValues)
                    .anyMatch(fValue ->
                        Arrays.stream(fValue.getLanguageItem()).anyMatch(languageItem -> languageItem.getTitle().equals(value))
                    );
                if (!matchFound && (extraEntries == null || Arrays.stream(extraEntries).noneMatch(ev -> ev.equals(value)))) {
                    notFoundValues.add(value);
                }
            }
            notFoundValues
                .stream()
                .map(v -> new AlternativesProposer(v, apiFieldName))
                .forEach(ap -> {
                    if (exactMatch) {
                        result.errors.add(
                            String.format("The field %s contains an invalid entry '%s'. %s", fieldName, ap.getTerm(), ap.toString()).trim()
                        );
                    } else {
                        result.warnings.add(
                            String.format("The field %s contains an unknown entry '%s'. %s", fieldName, ap.getTerm(), ap.toString()).trim()
                        );
                    }
                });
        }
    }

    private void specialFieldHandling(String[] values, String fieldName, ValidationResultDTO result) {
        // special handling of language field:
        switch (fieldName) {
            case "language":
                final List<String> supportedLanguageCodes = vocabularyRepository.getSupportedLanguageCodes();
                for (String languageCode : values) {
                    if (!supportedLanguageCodes.contains(languageCode)) {
                        result.errors.add(String.format("The language code '%s' is not yet supported.", languageCode));
                    }
                }
                break;
            case "somethingElse":
                break;
            default:
            // just ignore

        }
    }

    /**
     * just an internal class to propose an alternative String, for unknown entries.
     *
     * @author Micha
     *
     */

    protected class AlternativesProposer {

        String term;
        String fieldName;

        public AlternativesProposer(String term, String fieldName) {
            super();
            this.term = term;
            this.fieldName = fieldName;
        }

        private String findAlternative() {
            final VocabularyItem[] items = vocabularyRepository.getVocabularyItemsFor(fieldName);
            final Optional<String> alternativeO = proposeAlternative(term, items, new String[] {});
            // Warning: ld.getDistance returns a value between 0 (far away) and 1 (equal)!
            return alternativeO.filter(alternative -> ld.getDistance(alternative, term) >= MAXIMAL_DISTANCE).orElse(null);
        }

        @Override
        public String toString() {
            String alternative = findAlternative();
            return alternative == null ? "" : "Did you mean '" + alternative + "' instead of '" + term + "'?";
        }

        public String getTerm() {
            return term;
        }
    }

    protected AlternativesProposer createAlternativesProposer(String term, String fieldName) {
        return new AlternativesProposer(term, fieldName);
    }

    public Optional<String> proposeAlternative(String value, VocabularyItem[] fieldValues, String[] extraEntries) {
        final Optional<VocabularyItem> minVocabularyItem = Arrays
            .stream(fieldValues)
            .max((fv1, fv2) -> (int) ((getLevenSteinDistance(fv1, value) - getLevenSteinDistance(fv2, value)) * 100.0f));
        final Optional<VocabularyItem.LanguageItem> minLanguageItem = minVocabularyItem.flatMap(fv ->
            Arrays
                .stream(fv.getLanguageItem())
                .max((li1, li2) -> (int) ((ld.getDistance(li1.getTitle(), value) - ld.getDistance(li2.getTitle(), value)) * 100.0f))
        );
        return minLanguageItem.map(LanguageItem::getTitle);
    }

    @SuppressWarnings("PMD.FieldDeclarationsShouldBeAtStartOfClass")
    private final LevenshteinDistance ld = new LevenshteinDistance();

    public float getLevenSteinDistance(VocabularyItem vi, String s) {
        return Arrays
            .stream(vi.getLanguageItem())
            .map(LanguageItem::getTitle)
            .map(title -> (ld.getDistance(title, s)))
            .max((f1, f2) -> (int) ((f1 - f2) * 100))
            .orElse(Float.MAX_VALUE);
    }

    public VocabularyServiceConfig readVocabularyConfig() throws IOException {
        try (InputStream configStream = this.getClass().getResourceAsStream("vocabularyServiceConfig.json")) {
            ObjectMapper mapper = JsonMapper
                .builder()
                .configure(DeserializationFeature.READ_ENUMS_USING_TO_STRING, true)
                .configure(MapperFeature.ACCEPT_CASE_INSENSITIVE_ENUMS, true)
                .configure(DeserializationFeature.ACCEPT_SINGLE_VALUE_AS_ARRAY, true)
                .configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, true)
                .build();
            return mapper.readValue(configStream, VocabularyServiceConfig.class);
        }
    }

    public static class VocabularyServiceConfig {

        private VocabularySetting[] config;

        public enum RequiredEnum {
            REQUIRED,
            REQUIRED_ON_TOP_LEVEL,
            OPTIONAL,
        }

        public static class VocabularySetting {

            private final Logger log = LoggerFactory.getLogger(VocabularySetting.class);

            private String property;
            private String type;
            private String oeResourceType;
            private RequiredEnum required;
            private boolean exactMatch = false;
            private Method getter;
            private String[] extraEntries = {};

            public String getProperty() {
                return property;
            }

            @SuppressWarnings("PMD.NullAssignment")
            public void setProperty(String property) {
                this.property = property;
                getter = null;
            }

            public String getType() {
                return type;
            }

            public void setType(String type) {
                this.type = type.strip();
            }

            @JsonGetter("OEResourceType")
            public String getOEResourceType() {
                return oeResourceType;
            }

            @JsonSetter("OEResourceType")
            public void setOEResourceType(String oEResourceType) {
                oeResourceType = oEResourceType;
            }

            public RequiredEnum getRequired() {
                return required;
            }

            public void setRequired(RequiredEnum required) {
                this.required = required;
            }

            public Method getGetter() {
                if (getter == null) {
                    try {
                        getter = new PropertyDescriptor(property, UserProvidedMetadataDTO.class).getReadMethod();
                        if (!getter.getReturnType().getCanonicalName().equals(getType())) {
                            log.error("getter for {} has wrong type: expected {}", property, getType());
                            return null;
                        }
                    } catch (IntrospectionException e) {
                        log.error("Cannot find getter for {}", property, e);
                    }
                }

                return getter;
            }

            public String[] getExtraEntries() {
                return extraEntries.clone();
            }

            public void setExtraEntries(String[] extraEntries) {
                this.extraEntries = extraEntries.clone();
            }

            @Override
            public String toString() {
                return "VocabularySetting: for " + property + "(" + type + "/" + oeResourceType + ")";
            }

            public boolean isExactMatch() {
                return exactMatch;
            }

            public void setExactMatch(boolean exactMatch) {
                this.exactMatch = exactMatch;
            }
        }

        @SuppressWarnings("PMD.MethodReturnsInternalArray")
        protected VocabularySetting[] getConfig() {
            return config;
        }

        @SuppressWarnings("PMD.MethodReturnsInternalArray")
        protected Optional<VocabularySetting> getConfigSettingByOEResourceType(String oeResourceType) {
            return Arrays
                .stream(config)
                .filter(s -> s.getOEResourceType() != null && s.getOEResourceType().equals(oeResourceType))
                .findAny();
        }

        @SuppressWarnings("PMD.ArrayIsStoredDirectly")
        protected void setConfig(VocabularySetting[] config) {
            this.config = config;
        }
    }

    public static class ValidationResultDTO {

        List<String> errors = new ArrayList<>();
        List<String> warnings = new ArrayList<>();

        public List<String> getErrors() {
            return errors;
        }

        public void addError(String error) {
            if (errors == null) {
                errors = new ArrayList<>();
            }
            errors.add(error);
        }

        public void setErrors(List<String> errors) {
            this.errors = errors;
        }

        public List<String> getWarnings() {
            return warnings;
        }

        public void addWarning(String warning) {
            if (warnings == null) {
                warnings = new ArrayList<>();
            }
            warnings.add(warning);
        }

        public void setWarnings(List<String> warnings) {
            this.warnings = warnings;
        }
    }

    /**
     * for testing purposes only
     *
     * @return
     */
    protected VocabularyServiceConfig getConfig() {
        return config;
    }
}
