package at.ac.uibk.gitsearch.web.rest;

import at.ac.uibk.gitsearch.service.GitlabService;

import com.fasterxml.jackson.core.JsonFactory;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.ObjectReader;
import com.fasterxml.jackson.databind.PropertyNamingStrategies;

import java.util.List;

import javax.annotation.PostConstruct;
import javax.validation.Valid;

import org.apache.http.auth.AuthenticationException;
import org.gitlab4j.api.GitLabApiException;
import org.gitlab4j.api.models.Group;
import org.gitlab4j.api.systemhooks.SystemHookEvent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.ResponseEntity;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

@RestController
@RequestMapping("/api/gitlab")
public class GitlabResource {

    @SuppressWarnings("unused")
    private static final Logger log = LoggerFactory.getLogger(GitlabResource.class);

    private final GitlabService gitlabService;

    protected ObjectReader jsonReaderForGitLabEvents;

    @PostConstruct
    public void setUpJsonMapperForGitLabEvents() {
        ObjectMapper jsonMapperForGitLabEvents = new ObjectMapper(new JsonFactory());
        jsonMapperForGitLabEvents.findAndRegisterModules();
        jsonMapperForGitLabEvents.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
        jsonMapperForGitLabEvents.setPropertyNamingStrategy(PropertyNamingStrategies.SNAKE_CASE);
        jsonReaderForGitLabEvents = jsonMapperForGitLabEvents.readerFor(SystemHookEvent.class);
    }

    public GitlabResource(GitlabService gitlabService) {
        this.gitlabService = gitlabService;
    }

    /**
     * GET /gitlab/groups : Retrieve Gitlab groups of user
     * @return List of Gitlab groups
     * @throws GitLabApiException if an error occurs
     */
    @GetMapping("/groups")
    public ResponseEntity<List<Group>> getUserGroups() throws GitLabApiException {
        try {
            return ResponseEntity.ok(gitlabService.getAllGroupsWithAtLeastMaintainerRole());
        } catch (AuthenticationException ae) {
            return ResponseEntity.status(401).body(null);
        }
    }

    /**
     * GET /gitlab/subgroups/{group} : Retrieve subgroups of group
     * @param groupId ID of group to retrieve its subgroups
     * @return List of subgroups of given group
     * @throws GitLabApiException if an error occurs
     */
    @GetMapping("/subgroups/{groupId}")
    public ResponseEntity<List<Group>> getSubGroups(@PathVariable("groupId") int groupId) throws GitLabApiException {
        return ResponseEntity.ok(gitlabService.getSubGroups(groupId));
    }

    /**
     * processes a gitlab JSON event.
     * Unfortunatelly gitlab use Snake Case. This is not correctly recognized by the
     * standard REST service json parser.
     * Therefore we have to parse it manually.
     *
     * @param gitLabEvent
     * @throws JsonProcessingException
     */
    @PostMapping("/eventListener")
    public void gitEventListener(@Valid @RequestBody String gitLabEventString) throws JsonProcessingException {
        // there may be events that cannot be processed
        if (gitLabEventString.contains("event_name")) {
            log.info("\"event_name\" missing: Will not process {}", gitLabEventString);
            return;
        }

        SystemHookEvent gitLabEvent = jsonReaderForGitLabEvents.readValue(gitLabEventString);
        gitlabService.handleGitLabEvent(gitLabEvent);
    }
}
